;; Generated package description from gptel.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "gptel" "0.9.9.5.0.20260713.61" "Interact with ChatGPT or other LLMs" '((emacs "27.1") (transient "0.7.8") (compat "30.1.0.0")) :commit "443885f96d2002d29ab2f76fca4ef43de569355a" :authors '(("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com")) :maintainer '("Karthik Chikmagalur" . "karthik.chikmagalur@gmail.com") :keywords '("convenience" "tools") :url "https://github.com/karthink/gptel")
